=== modified file 'NEWS'
--- NEWS	2009-09-30 05:58:45 +0000
+++ NEWS	2009-10-01 00:56:56 +0000
@@ -22,6 +22,14 @@
   ``keyboard-interactive`` auth but not ``password`` auth when using
   Paramiko.   (Andrew Bennetts, #433846)
 
+* Occasional IndexError on renamed files have been fixed. Operations that
+  set a full inventory in the working tree will now go via the
+  apply_inventory_delta code path which is simpler and easier to
+  understand than dirstates set_state_from_inventory method. This may
+  have a small performance impact on operations built on _write_inventory,
+  but such operations are already doing full tree scans, so no radical
+  performance change should be observed. (Robert Collins, #403322)
+
 * When a file kind becomes unversionable after being added, a sensible
   error will be shown instead of a traceback. (Robert Collins, #438569)
 

=== modified file 'bzrlib/dirstate.py'
--- bzrlib/dirstate.py	2009-09-30 05:58:45 +0000
+++ bzrlib/dirstate.py	2009-10-01 00:56:56 +0000
@@ -1324,7 +1324,7 @@
                 key = (dirname_utf8, basename, file_id)
                 minikind = DirState._kind_to_minikind[inv_entry.kind]
                 if minikind == 't':
-                    fingerprint = inv_entry.reference_revision
+                    fingerprint = inv_entry.reference_revision or ''
                 else:
                     fingerprint = ''
                 insertions[file_id] = (key, minikind, inv_entry.executable,

=== modified file 'bzrlib/workingtree_4.py'
--- bzrlib/workingtree_4.py	2009-08-25 04:43:21 +0000
+++ bzrlib/workingtree_4.py	2009-09-30 21:38:49 +0000
@@ -1267,9 +1267,17 @@
         if self._dirty:
             raise AssertionError("attempting to write an inventory when the "
                 "dirstate is dirty will lose pending changes")
-        self.current_dirstate().set_state_from_inventory(inv)
-        self._make_dirty(reset_inventory=False)
-        if self._inventory is not None:
+        had_inventory = self._inventory is not None
+        # Setting self._inventory = None forces the dirstate to regenerate the
+        # working inventory. We do this because self.inventory may be inv, or
+        # may have been modified, and either case would prevent a clean delta
+        # being created.
+        self._inventory = None
+        # generate a delta,
+        delta = inv._make_delta(self.inventory)
+        # and apply it.
+        self.apply_inventory_delta(delta)
+        if had_inventory:
             self._inventory = inv
         self.flush()
 

